from qiskit import QuantumCircuit, transpile

import matplotlib.pyplot as plt
import json
import importlib
import signal
import argparse
import numpy as np

from typing import Union
from os import path, mkdir, remove

from mqt.bench.utils import utils
from mqt.bench.benchmarks import (
    shor,
    hhl,
)
from mqt.bench.benchmarks.qiskit_application_optimization import routing, tsp
from mqt.bench.benchmarks.qiskit_application_finance import (
    pricingcall,
    pricingput,
)
from mqt.bench.benchmarks.qiskit_application_nature import groundstate, excitedstate


def init_module_paths():
    global benchmarks_module_paths_dict
    benchmarks_module_paths_dict = {
        "ae": "mqt.bench.benchmarks.ae",
        "dj": "mqt.bench.benchmarks.dj",
        "grover": "mqt.bench.benchmarks.grover",
        "ghz": "mqt.bench.benchmarks.ghz",
        "graphstate": "mqt.bench.benchmarks.graphstate",
        "portfolioqaoa": "mqt.bench.benchmarks.qiskit_application_finance.portfolioqaoa",
        "portfoliovqe": "mqt.bench.benchmarks.qiskit_application_finance.portfoliovqe",
        "qaoa": "mqt.bench.benchmarks.qaoa",
        "qft": "mqt.bench.benchmarks.qft",
        "qftentangled": "mqt.bench.benchmarks.qftentangled",
        "qgan": "mqt.bench.benchmarks.qiskit_application_ml.qgan",
        "qpeexact": "mqt.bench.benchmarks.qpeexact",
        "qpeinexact": "mqt.bench.benchmarks.qpeinexact",
        "qwalk": "mqt.bench.benchmarks.qwalk",
        "realamprandom": "mqt.bench.benchmarks.realamprandom",
        "su2random": "mqt.bench.benchmarks.su2random",
        "twolocalrandom": "mqt.bench.benchmarks.twolocalrandom",
        "vqe": "mqt.bench.benchmarks.vqe",
        "wstate": "mqt.bench.benchmarks.wstate",
    }


def create_benchmarks_from_config(cfg_path: str):
    init_module_paths()
    characteristics = []

    with open(cfg_path, "r") as jsonfile:
        cfg = json.load(jsonfile)
        print("Read config successful")

    # global seetings
    global save_png
    save_png = cfg["save_png"]
    global save_hist
    save_hist = cfg["save_hist"]
    global max_depth
    max_depth = cfg["max_depth"]
    global timeout
    timeout = cfg["timeout"]

    global qasm_output_folder
    qasm_output_folder = utils.get_qasm_output_path()
    if not path.isdir(qasm_output_folder):
        mkdir(qasm_output_folder)

    for benchmark in cfg["benchmarks"]:
        print(benchmark["name"])
        characteristics.extend(generate_benchmark(benchmark))
    return characteristics


def benchmark_generation_watcher(func, args):
    class TimeoutException(Exception):  # Custom exception class
        pass

    def timeout_handler(signum, frame):  # Custom signal handler
        raise TimeoutException("TimeoutException")

    # Change the behavior of SIGALRM
    signal.signal(signal.SIGALRM, timeout_handler)

    signal.alarm(timeout)
    try:
        filename, depth, num_qubits = func(*args)
    except Exception as e:
        # print("Calculation/Generation exceeded timeout limit for ", func, args[1:])
        print("Exception: ", e, func, args[1:])
        if func == utils.get_indep_level:
            qc = args[0]
            num_qubits = args[1]
            filename_indep = qc.name + "_t-indep_" + str(num_qubits)
            file_path = qasm_output_folder + filename_indep + ".qasm"

            if path.isfile(file_path):
                remove(file_path)
                # print("removed file: ", path)

        elif func == utils.get_native_gates_level:
            qc = args[0]
            gate_set_name = args[2]
            opt_level = args[3]
            num_qubits = args[4]

            filename_nativegates = (
                qc.name
                + "_nativegates_"
                + gate_set_name
                + "_opt"
                + str(opt_level)
                + "_"
                + str(num_qubits)
            )

            file_path = qasm_output_folder + filename_nativegates + ".qasm"
            if path.isfile(file_path):
                remove(file_path)

        elif func == utils.get_mapped_level:
            qc = args[0]
            gate_set_name_mapped = args[2]
            opt_level = args[3]
            num_qubits = args[4]
            smallest_arch = args[5]
            if smallest_arch:
                gate_set_name_mapped += "-s"
            else:
                gate_set_name_mapped += "-b"

            filename_mapped = (
                qc.name
                + "_mapped_"
                + gate_set_name_mapped
                + "_opt"
                + str(opt_level)
                + "_"
                + str(num_qubits)
            )

            file_path = qasm_output_folder + filename_mapped + ".qasm"
            if path.isfile(file_path):
                remove(file_path)

        return False
    else:
        # Reset the alarm
        signal.alarm(0)

    if depth > max_depth:
        print("Depth of generated circuit is too large: ", depth)
        return False

    return filename, depth, num_qubits


def qc_creation_watcher(func, args):
    class TimeoutException(Exception):  # Custom exception class
        pass

    def timeout_handler(signum, frame):  # Custom signal handler
        raise TimeoutException

    # Change the behavior of SIGALRM
    signal.signal(signal.SIGALRM, timeout_handler)

    signal.alarm(timeout)
    try:
        qc, num_qubits, file_precheck = func(*args)
    except TimeoutException:
        print("Calculation/Generation exceeded timeout limit for ", func, args[1:])
        return False
    except Exception as e:
        print("Something else went wrong: ", e)
        return False
    else:
        # Reset the alarm
        signal.alarm(0)

    return qc, num_qubits, file_precheck


def generate_benchmark(benchmark):
    characteristics = []
    if benchmark["include"]:
        if benchmark["name"] == "grover" or benchmark["name"] == "qwalk":
            for anc_mode in benchmark["ancillary_mode"]:
                for n in range(
                    benchmark["min_qubits"],
                    benchmark["max_qubits"],
                    benchmark["stepsize"],
                ):
                    res_qc_creation = qc_creation_watcher(
                        create_scalable_qc, [benchmark, n, anc_mode]
                    )
                    if not res_qc_creation:
                        break
                    res = generate_circuits_on_all_levels(*res_qc_creation)
                    if len(res) == 0:
                        break
                    characteristics.extend(res)

        elif benchmark["name"] == "shor":
            for choice in benchmark["instances"]:
                res_qc_creation = qc_creation_watcher(create_shor_qc, [choice])
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)

        elif benchmark["name"] == "hhl":
            for i in range(benchmark["min_index"], benchmark["max_index"]):
                res_qc_creation = qc_creation_watcher(create_hhl_qc, [i])
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)

        elif benchmark["name"] == "routing":
            for nodes in range(benchmark["min_nodes"], benchmark["max_nodes"]):
                res_qc_creation = qc_creation_watcher(create_routing_qc, [nodes])
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)

        elif benchmark["name"] == "tsp":
            for nodes in range(benchmark["min_nodes"], benchmark["max_nodes"]):
                res_qc_creation = qc_creation_watcher(create_tsp_qc, [nodes])
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)

        elif benchmark["name"] == "groundstate":
            for choice in benchmark["instances"]:
                res_qc_creation = qc_creation_watcher(create_groundstate_qc, [choice])
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)

        elif benchmark["name"] == "excitedstate":
            for choice in benchmark["instances"]:
                res_qc_creation = qc_creation_watcher(create_excitedstate_qc, [choice])
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)

        elif benchmark["name"] == "pricingcall":
            for nodes in range(
                benchmark["min_uncertainty"], benchmark["max_uncertainty"]
            ):
                res_qc_creation = qc_creation_watcher(create_pricingcall_qc, [nodes])
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)

        elif benchmark["name"] == "pricingput":
            for nodes in range(
                benchmark["min_uncertainty"], benchmark["max_uncertainty"]
            ):
                res_qc_creation = qc_creation_watcher(create_pricingput_qc, [nodes])
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)
        else:
            for n in range(
                benchmark["min_qubits"], benchmark["max_qubits"], benchmark["stepsize"]
            ):

                # res_qc_creation == qc, num_qubits, file_precheck
                # res == filename, depth, num_qubits
                res_qc_creation = qc_creation_watcher(
                    create_scalable_qc, [benchmark, n]
                )
                if not res_qc_creation:
                    break
                res = generate_circuits_on_all_levels(*res_qc_creation)
                if len(res) == 0:
                    break
                characteristics.extend(res)

    return characteristics


def generate_circuits_on_all_levels(qc, num_qubits, file_precheck):
    characteristics = []
    # filename_algo, depth = generate_algo_level_circuit(qc, n, save_png, save_hist)
    # characteristics.append([filename_algo, n, depth])

    res_t_indep = generate_target_indep_level_circuit(
        qc, num_qubits, save_png, save_hist, file_precheck
    )

    if res_t_indep:
        characteristics.extend(res_t_indep)
    else:
        return characteristics

    res_t_dep = generate_target_dep_level_circuit(
        qc, num_qubits, save_png, save_hist, file_precheck
    )

    if res_t_dep:
        characteristics.extend(res_t_dep)
    else:
        return characteristics

    return characteristics


def generate_algo_level_circuit(
    qc: QuantumCircuit, num_qubits: int, save_png, save_hist
):
    characteristics = []
    res = benchmark_generation_watcher(
        utils.handle_algorithm_level, [qc, num_qubits, save_png, save_hist]
    )
    characteristics.append(res)

    if res:
        characteristics.append(res)
        return characteristics
    else:
        return False


def generate_target_indep_level_circuit(
    qc: QuantumCircuit, num_qubits: int, save_png, save_hist, file_precheck
):
    characteristics = []

    res = benchmark_generation_watcher(
        utils.get_indep_level,
        [qc, num_qubits, save_png, save_hist, file_precheck],
    )
    if res:
        characteristics.append(res)
        return characteristics
    else:
        return False


def generate_target_dep_level_circuit(
    qc: QuantumCircuit, num_qubits: int, save_png, save_hist, file_precheck
):
    characteristics = []

    ibm_native_gates = utils.FakeMontreal().configuration().basis_gates
    rigetti_native_gates = utils.get_rigetti_native_gates()
    gate_sets = [(ibm_native_gates, "ibm"), (rigetti_native_gates, "rigetti")]

    for gate_set, gate_set_name in gate_sets:
        try:
            for opt_level in range(4):

                # Creating the circuit on target-dependent: native gates level

                res = benchmark_generation_watcher(
                    utils.get_native_gates_level,
                    [
                        qc,
                        gate_set,
                        gate_set_name,
                        opt_level,
                        num_qubits,
                        save_png,
                        save_hist,
                        file_precheck,
                    ],
                )
                if res:
                    characteristics.append(res)
                else:
                    break
                n_actual = res[2]

                # Creating the circuit on target-dependent: mapped level for both mapping schemes
                res = benchmark_generation_watcher(
                    utils.get_mapped_level,
                    [
                        qc,
                        gate_set,
                        gate_set_name,
                        opt_level,
                        n_actual,
                        True,
                        save_png,
                        save_hist,
                        file_precheck,
                    ],
                )

                if res:
                    characteristics.append(res)
                else:
                    break
                res = benchmark_generation_watcher(
                    utils.get_mapped_level,
                    [
                        qc,
                        gate_set,
                        gate_set_name,
                        opt_level,
                        n_actual,
                        False,
                        save_png,
                        save_hist,
                        file_precheck,
                    ],
                )

                if res:
                    characteristics.append(res)
                else:
                    break
        except Exception as e:
            print(
                "\n Problem occured in inner loop: ",
                qc.name,
                num_qubits,
                gate_set_name,
                e,
            )

    return characteristics


def create_scalable_qc(benchmark, num_qubits, ancillary_mode=None):
    file_precheck = True
    try:
        # Creating the circuit on Algorithmic Description level
        lib = importlib.import_module(benchmarks_module_paths_dict[benchmark["name"]])
        if benchmark["name"] == "grover" or benchmark["name"] == "qwalk":
            qc = lib.create_circuit(num_qubits, ancillary_mode=ancillary_mode)
            qc.name = qc.name + "-" + ancillary_mode
            file_precheck = False

        else:
            qc = lib.create_circuit(num_qubits)

        n = qc.num_qubits
        return qc, n, file_precheck

    except Exception as e:
        print("\n Problem occured in outer loop: ", benchmark, num_qubits, e)


def create_shor_qc(choice: str):
    instances = {
        "xsmall": [9, 4],  # 18 qubits
        "small": [15, 4],  # 18 qubits
        "medium": [821, 4],  # 42 qubits
        "large": [11777, 4],  # 58 qubits
        "xlarge": [201209, 4],  # 74 qubits
    }

    try:
        qc = shor.create_circuit(instances[choice][0], instances[choice][1])
        return qc, qc.num_qubits, False

    except Exception as e:
        print(
            "\n Problem occured in outer loop: ", "create_shor_benchmarks: ", choice, e
        )


def create_hhl_qc(index: int):
    # index is not the number of qubits in this case
    try:
        # Creating the circuit on Algorithmic Description level
        qc = hhl.create_circuit(index)
        return qc, qc.num_qubits, False

    except Exception as e:
        print("\n Problem occured in outer loop: ", "create_hhl_benchmarks", index, e)


def create_routing_qc(nodes: int):
    try:
        # Creating the circuit on Algorithmic Description level
        qc = routing.create_circuit(nodes, 2)
        return qc, qc.num_qubits, False

    except Exception as e:
        print(
            "\n Problem occured in outer loop: ", "create_routing_benchmarks", nodes, e
        )


def create_tsp_qc(nodes: int):
    try:
        # Creating the circuit on Algorithmic Description level
        qc = tsp.create_circuit(nodes)
        return qc, qc.num_qubits, False

    except Exception as e:
        print("\n Problem occured in outer loop: ", "create_tsp_benchmarks", nodes, e)


def create_groundstate_qc(choice: str):
    molecule = utils.get_molecule(choice)

    try:
        qc = groundstate.create_circuit(molecule)
        qc.name = qc.name + "-" + choice
        return qc, qc.num_qubits, False

    except Exception as e:
        print(
            "\n Problem occured in outer loop: ",
            "create_groundstate_benchmarks",
            choice,
            e,
        )


def create_excitedstate_qc(choice: str):
    molecule = utils.get_molecule(choice)

    try:
        qc = excitedstate.create_circuit(molecule)
        qc.name = qc.name + "-" + choice
        return qc, qc.num_qubits, False

    except Exception as e:
        print(
            "\n Problem occured in outer loop: ",
            "create_excitedstate_benchmarks",
            choice,
            e,
        )
        return False


def create_pricingcall_qc(num_uncertainty: int):
    # num_options is not the number of qubits in this case
    try:
        # Creating the circuit on Algorithmic Description level
        qc = pricingcall.create_circuit(num_uncertainty)
        return qc, qc.num_qubits, False

    except Exception as e:
        print(
            "\n Problem occured in outer loop: ",
            "create_pricingcall_benchmarks",
            num_uncertainty,
            e,
        )


def create_pricingput_qc(num_uncertainty: int):
    # num_uncertainty is not the number of qubits in this case
    try:
        # Creating the circuit on Algorithmic Description level
        qc = pricingput.create_circuit(num_uncertainty)
        return qc, qc.num_qubits, False

    except Exception as e:
        print(
            "\n Problem occured in outer loop: ",
            "create_pricingput_benchmarks",
            num_uncertainty,
            e,
        )


def save_benchmark_hist(characteristics):
    characteristics = np.array(characteristics)
    plt.scatter(
        x=characteristics[:, 2].astype(int), y=characteristics[:, 1].astype(int)
    )
    plt.yscale("log")
    plt.title("Depth and Width of generated Benchmarks")
    plt.xlabel("# of Qubits")
    plt.ylabel("Circuit Depth")
    plt.savefig("benchmark_histogram")


def get_one_benchmark(
    benchmark_name: str,
    level: Union[str, int],
    circuit_size: int = None,
    benchmark_instance_name: str = None,
    opt_level: int = 0,
    gate_set_name: str = "ibm",
    smallest_fitting_arch: bool = True,
):
    """Returns one benchmark as a Qiskit::QuantumCircuit Object.

    Keyword arguments:
    benchmark_name -- name of the to be generated benchmark
    level -- Choice of level, either as a string ("alg", "indep", "gates" or "mapped") or as a number between 0-3 where
    circuit_size -- Input for the benchmark creation, in most cases this is equal to the qubit number
    benchmark_instance_name -- Input selection for some benchmarks, namely "groundstate", "excitedstate" and "shor"
    0 corresponds to "alg" level and 3 to "mapped" level
    opt_level -- Level of optimization (relevant to "gates" and "mapped" levels)
    gate_set_name -- Either "ibm" or "rigetti"
    smallest_fitting_arch -- True->Smallest architecture is selected, False->Biggest one is selected (ibm: 127 qubits,
    rigetti: 80 qubits)


    Return values:
    QuantumCircuit -- Representing the benchmark with the selected options
    """
    init_module_paths()
    ibm_native_gates = utils.FakeMontreal().configuration().basis_gates
    rigetti_native_gates = utils.get_rigetti_native_gates()

    if gate_set_name and "ibm" in gate_set_name:
        gate_set = ibm_native_gates
    elif gate_set_name and "rigetti" in gate_set_name:
        gate_set = rigetti_native_gates
    else:
        gate_set = None

    if "grover" in benchmark_name or "qwalk" in benchmark_name:
        if "noancilla" in benchmark_name:
            anc_mode = "noancilla"
        elif "v-chain" in benchmark_name:
            anc_mode = "v-chain"

        short_name = benchmark_name.split("-")[0]
        lib = importlib.import_module(benchmarks_module_paths_dict[short_name])
        qc = lib.create_circuit(circuit_size, ancillary_mode=anc_mode)
        qc.name = qc.name + "-" + anc_mode

    elif benchmark_name == "shor":
        instances = {
            "xsmall": [9, 4],  # 18 qubits
            "small": [15, 4],  # 18 qubits
            "medium": [821, 4],  # 42 qubits
            "large": [11777, 4],  # 58 qubits
            "xlarge": [201209, 4],  # 74 qubits
        }

        qc = shor.create_circuit(*instances[benchmark_instance_name])

    elif benchmark_name == "hhl":
        qc = hhl.create_circuit(circuit_size)

    elif benchmark_name == "routing":
        qc = routing.create_circuit(circuit_size)

    elif benchmark_name == "tsp":
        qc = tsp.create_circuit(circuit_size)

    elif benchmark_name == "groundstate":
        molecule = utils.get_molecule(benchmark_instance_name)
        qc = groundstate.create_circuit(molecule)

    elif benchmark_name == "excitedstate":
        molecule = utils.get_molecule(benchmark_instance_name)
        qc = excitedstate.create_circuit(molecule)

    elif benchmark_name == "pricingcall":
        qc = pricingcall.create_circuit(circuit_size)

    elif benchmark_name == "pricingput":
        qc = pricingput.create_circuit(circuit_size)

    else:
        lib = importlib.import_module(benchmarks_module_paths_dict[benchmark_name])
        qc = lib.create_circuit(circuit_size)

    if level == "alg" or level == 0:
        return qc

    elif level == "indep" or level == 1:

        qc_indep = transpile(qc, basis_gates=utils.get_openqasm_gates())
        return qc_indep

    elif level == "nativegates" or level == 2:
        qc_gates = utils.get_compiled_circuit_with_gateset(
            qc=qc, opt_level=opt_level, basis_gates=gate_set
        )
        return qc_gates

    elif level == "mapped" or level == 3:
        c_map, backend_name, gate_set_name_mapped, c_map_found = utils.select_c_map(
            gate_set_name, smallest_fitting_arch, circuit_size
        )
        if c_map_found:
            qc_mapped = utils.get_compiled_circuit_with_gateset(
                qc=qc, opt_level=opt_level, basis_gates=gate_set, c_map=c_map
            )
            return qc_mapped
        else:
            return print("No Hardware Architecture available for that config.")
    else:
        return print("Level specification was wrong.")


if __name__ == "__main__":
    init_module_paths()

    parser = argparse.ArgumentParser(description="Create Configuration")
    parser.add_argument(
        "--file-name", type=str, help="optional filename", default="./config.json"
    )

    args = parser.parse_args()

    print(args.file_name)
    characteristics = create_benchmarks_from_config(args.file_name)
    save_benchmark_hist(characteristics)



# Code from https://qiskit.org/documentation/machine-learning/tutorials/04_qgans_for_loading_random_distributions.html

import numpy as np
from qiskit import Aer
from qiskit.circuit.library import TwoLocal
from qiskit.utils import QuantumInstance, algorithm_globals


def create_circuit(num_qubits: int):
    """Returns a quantum circuit implementing Quantum Generative Adversarial Networks algorithm.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """
    try:
        from qiskit_finance.circuit.library import UniformDistribution
    except:
        print("Please install qiskit_finance.")
        return None

    try:
        from qiskit_machine_learning.algorithms import NumPyDiscriminator, QGAN
    except:
        print("Please install qiskit_machine_learning.")
        return None

    seed = 10
    np.random.seed = seed
    algorithm_globals.random_seed = seed

    # Number training data samples
    n = 100

    # Load data samples from log-normal distribution with mean=1 and standard deviation=1
    mu = 1
    sigma = 1
    real_data = np.random.lognormal(mean=mu, sigma=sigma, size=n)

    # Set the data resolution
    # Set upper and lower data values as list of k min/max data values [[min_0,max_0],...,[min_k-1,max_k-1]]
    upper_bound_value = 2**num_qubits - 1
    bounds = np.array([0.0, upper_bound_value])
    # Set number of qubits per data dimension as list of k qubit values[#q_0,...,#q_k-1]
    num_qubits = [num_qubits]

    # Set number of training epochs
    # Note: The algorithm's runtime can be shortened by reducing the number of training epochs.
    num_epochs = 5
    # Batch size
    batch_size = 10

    # Initialize qGAN
    qgan = QGAN(
        real_data, bounds, num_qubits, batch_size, num_epochs, snapshot_dir=None
    )
    qgan.seed = 10
    # Set quantum instance to run the quantum generator
    quantum_instance = QuantumInstance(
        backend=Aer.get_backend("aer_simulator"),
        seed_transpiler=seed,
        seed_simulator=seed,
        shots=1024,
    )

    # Set an initial state for the generator circuit
    init_dist = UniformDistribution(sum(num_qubits))

    # Set the ansatz circuit
    ansatz = TwoLocal(
        int(np.sum(num_qubits)), "ry", "cz", reps=1
    )  # entanglement=entangler_map,

    # You can increase the number of training epochs and use random initial parameters.
    init_params = np.random.rand(ansatz.num_parameters_settable) * 2 * np.pi

    # Set generator circuit by adding the initial distribution infront of the ansatz
    g_circuit = ansatz.compose(init_dist, front=True)

    # Set quantum generator
    qgan.set_generator(generator_circuit=g_circuit, generator_init_params=init_params)
    # The parameters have an order issue that following is a temp. workaround
    qgan._generator._free_parameters = sorted(
        g_circuit.parameters, key=lambda p: p.name
    )
    # Set classical discriminator neural network
    discriminator = NumPyDiscriminator(len(num_qubits))
    qgan.set_discriminator(discriminator)

    result = qgan.run(quantum_instance)

    param_values = qgan.generator.parameter_values
    qc = qgan.generator.construct_circuit(params=param_values)
    qc.measure_all()
    qc.name = "qgan"

    return qc



from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit.library import GraphState

import networkx as nx


def create_circuit(num_qubits: int, degree: int = 2):
    """Returns a quantum circuit implementing a graph state.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    degree -- number of edges per node
    """

    q = QuantumRegister(num_qubits, "q")
    qc = QuantumCircuit(q, name="graphstate")

    g = nx.random_regular_graph(degree, num_qubits)
    a = nx.convert_matrix.to_numpy_array(g)
    qc.compose(GraphState(a), inplace=True)
    qc.measure_all()

    return qc



from qiskit.circuit.library import QFT
from qiskit import QuantumCircuit, QuantumRegister


def create_circuit(num_qubits: int):
    """Returns a quantum circuit implementing the Quantum Fourier Transform algorithm using entangled qubits.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """

    q = QuantumRegister(num_qubits, "q")
    qc = QuantumCircuit(q)
    qc.h(q[-1])
    for i in range(1, num_qubits):
        qc.cx(q[num_qubits - i], q[num_qubits - i - 1])

    qc.compose(QFT(num_qubits=num_qubits), inplace=True)

    qc.measure_all()
    qc.name = "qftentangled"

    return qc



from qiskit.circuit.library import TwoLocal
import numpy as np


def create_circuit(num_qubits: int):
    """Returns a quantum circuit implementing the TwoLocal ansatz with random parameter
    values.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """

    np.random.seed = 10
    qc = TwoLocal(num_qubits, "ry", "cx", entanglement="full", reps=3)
    num_params = qc.num_parameters
    qc = qc.bind_parameters(np.random.rand(num_params))
    qc.measure_all()
    qc.name = "twolocalrandom"

    return qc



from qiskit.algorithms.factorizers import Shor


def create_circuit(num_to_be_factorized: int, a: int = 2):
    """Returns a quantum circuit implementing the Shor's algorithm.

    Keyword arguments:
    num_to_be_factorized -- number which shall be factorized
    a -- any integer that satisfies 1 < a < num_to_be_factorized and gcd(a, num_to_be_factorized) = 1
    """

    qc = Shor().construct_circuit(num_to_be_factorized, a)
    qc.measure_all()
    qc.name = "shor_" + str(num_to_be_factorized) + "_" + str(a)

    return qc



from qiskit import QuantumCircuit, QuantumRegister, AncillaRegister
from qiskit.circuit.library import GroverOperator
from qiskit.algorithms import Grover
import numpy as np


def create_circuit(num_qubits: int, ancillary_mode: str = "noancilla"):
    """Returns a quantum circuit implementing Grover's algorithm.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    ancillary_mode -- defining the decomposition scheme
    """

    num_qubits = num_qubits - 1  # -1 because of the flag qubit
    q = QuantumRegister(num_qubits, "q")
    flag = AncillaRegister(1, "flag")

    state_preparation = QuantumCircuit(q, flag)
    state_preparation.h(q)
    state_preparation.x(flag)

    oracle = QuantumCircuit(q, flag)
    oracle.mcp(np.pi, q, flag)

    operator = GroverOperator(oracle, mcx_mode=ancillary_mode)
    iterations = Grover.optimal_num_iterations(1, num_qubits)

    num_qubits = (
        operator.num_qubits - 1
    )  # -1 because last qubit is "flag" qubit and already taken care of

    # num_qubits may differ now depending on the mcx_mode
    q2 = QuantumRegister(num_qubits, "q")
    qc = QuantumCircuit(q2, flag, name="grover")
    qc.compose(state_preparation, inplace=True)

    qc.compose(operator.power(iterations), inplace=True)
    qc.measure_all()

    return qc



# Code from https://qiskit.org/textbook/ch-algorithms/deutsch-jozsa.html

from qiskit import QuantumCircuit
import numpy as np


def dj_oracle(case, n):
    # plus one output qubit
    oracle_qc = QuantumCircuit(n + 1)

    if case == "balanced":
        np.random.seed = 10
        b_str = ""
        for _ in range(n):
            b = np.random.randint(0, 2)
            b_str = b_str + str(b)

        for qubit in range(len(b_str)):
            if b_str[qubit] == "1":
                oracle_qc.x(qubit)

        for qubit in range(n):
            oracle_qc.cx(qubit, n)

        for qubit in range(len(b_str)):
            if b_str[qubit] == "1":
                oracle_qc.x(qubit)

    if case == "constant":
        output = np.random.randint(2)
        if output == 1:
            oracle_qc.x(n)

    oracle_gate = oracle_qc.to_gate()
    oracle_gate.name = "Oracle"  # To show when we display the circuit
    return oracle_gate


def dj_algorithm(oracle, n):
    dj_circuit = QuantumCircuit(n + 1, n)

    dj_circuit.x(n)
    dj_circuit.h(n)

    for qubit in range(n):
        dj_circuit.h(qubit)

    dj_circuit.append(oracle, range(n + 1))

    for qubit in range(n):
        dj_circuit.h(qubit)

    dj_circuit.barrier()
    for i in range(n):
        dj_circuit.measure(i, i)

    return dj_circuit


def create_circuit(n: int, balanced: bool = True):
    """Returns a quantum circuit implementing the Deutsch-Josza algorithm.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    balanced -- True for a balanced and False for a constant oracle
    """

    if balanced:
        oracle_mode = "balanced"
    else:
        oracle_mode = "constant"
    n = n - 1  # because of ancilla qubit
    oracle_gate = dj_oracle(oracle_mode, n)
    qc = dj_algorithm(oracle_gate, n)
    qc.name = "dj"

    return qc



from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
from qiskit.circuit.library import QFT


def create_circuit(num_qubits: int):
    """Returns a quantum circuit implementing the Quantum Fourier Transform algorithm.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """

    q = QuantumRegister(num_qubits, "q")
    c = ClassicalRegister(num_qubits, "c")
    qc = QuantumCircuit(q, c, name="qft")
    qc.compose(QFT(num_qubits=num_qubits), inplace=True)
    qc.measure_all()

    return qc



# Code based on https://qiskit.org/textbook/ch-applications/hhl_tutorial.html

import numpy as np
from qiskit.algorithms.linear_solvers.matrices.tridiagonal_toeplitz import (
    TridiagonalToeplitz,
)
from qiskit.algorithms.linear_solvers.hhl import HHL


def create_circuit(num_qubits: int):
    """Returns a quantum circuit implementing the HHL algorithm for a specific example matrix.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """

    a = 1
    b = -1 / 3

    vector = np.array([1] + [0] * (2**num_qubits - 1))
    tridi_matrix = TridiagonalToeplitz(num_qubits, a, b)
    qc = HHL().solve(tridi_matrix, vector).state

    qc.name = "hhl"
    qc.measure_all()

    return qc



# Code from https://github.com/qiskit-community/qiskit-application-modules-demo-sessions/blob/main/qiskit-optimization/qiskit-optimization-demo.ipynb

from qiskit import Aer
from qiskit.utils import QuantumInstance
from qiskit_optimization.algorithms import MinimumEigenOptimizer
from qiskit.algorithms import QAOA
from qiskit.algorithms.optimizers import SLSQP

from mqt.bench.utils.utils import get_examplary_max_cut_qp


def create_circuit(num_qubits: int):
    """Returns a quantum circuit implementing the Quantum Approximation Optimization Algorithm for a specific max-cut
     example.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """

    qp = get_examplary_max_cut_qp(num_qubits)
    sim = QuantumInstance(
        backend=Aer.get_backend("aer_simulator"), shots=1024, seed_simulator=10
    )

    qaoa = QAOA(reps=2, optimizer=SLSQP(maxiter=25), quantum_instance=sim)
    qaoa_result = qaoa.compute_minimum_eigenvalue(qp.to_ising()[0])
    qc = qaoa.ansatz.bind_parameters(qaoa_result.optimal_point)

    qc.measure_all()
    qc.name = "qaoa"

    return qc



# Code from https://qiskit.org/documentation/finance/tutorials/00_amplitude_estimation.html

from qiskit.algorithms import AmplitudeEstimation

from mqt.bench.utils import utils


def create_circuit(num_qubits: int):
    """Returns a quantum circuit implementing Quantum Amplitude Estimation.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """

    ae = AmplitudeEstimation(
        num_eval_qubits=num_qubits - 1,  # -1 because of the to be estimated qubit
    )
    problem = utils.get_estimation_problem()

    qc = ae.construct_circuit(problem)
    qc.name = "ae"
    qc.measure_all()

    return qc



import random
import numpy as np

from fractions import Fraction

from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit
from qiskit.circuit.library import QFT


def create_circuit(num_qubits: int):
    """Returns a quantum circuit implementing the Quantum Phase Estimation algorithm for a phase which cannot be
    exactly estimated.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """

    num_qubits = num_qubits - 1  # because of ancilla qubit
    q = QuantumRegister(num_qubits, "q")
    psi = QuantumRegister(1, "psi")
    c = ClassicalRegister(num_qubits, "c")
    qc = QuantumCircuit(q, psi, c, name="qpeinexact")

    # get random n+1-bit string as target phase
    random.seed(10)
    theta = 0
    while theta == 0 or (theta & 1) == 0:
        theta = random.getrandbits(num_qubits + 1)
    lam = Fraction(0, 1)
    # print("theta : ", theta, "correspond to", theta / (1 << (n+1)), "bin: ")
    for i in range(num_qubits + 1):
        if theta & (1 << (num_qubits - i)):
            lam += Fraction(1, (1 << i))

    qc.x(psi)
    qc.h(q)

    for i in range(num_qubits):
        angle = (lam * (1 << i)) % 2
        if angle > 1:
            angle -= 2
        if angle != 0:
            qc.cp(angle * np.pi, psi, q[i])

    qc.compose(
        QFT(num_qubits=num_qubits, inverse=True),
        inplace=True,
        qubits=list(range(num_qubits)),
    )
    qc.barrier()
    qc.measure(q, c)

    return qc



# Code from https://qiskit.org/documentation/nature/tutorials/03_ground_state_solvers.html


from qiskit import Aer
from qiskit.utils import QuantumInstance
from qiskit.circuit.library import TwoLocal
from qiskit.algorithms import VQE
from qiskit.algorithms.optimizers import COBYLA


def create_circuit(molecule, basis: str = "sto3g"):
    """Returns a quantum circuit implementing Ground State Estimation.

    Keyword arguments:
    molecule -- Molecule for which the ground state shall be estimated.
    """

    try:
        from qiskit_nature.drivers import Molecule
        from qiskit_nature.drivers.second_quantization import (
            ElectronicStructureDriverType,
            ElectronicStructureMoleculeDriver,
        )
        from qiskit_nature.problems.second_quantization import (
            ElectronicStructureProblem,
        )
        from qiskit_nature.converters.second_quantization import QubitConverter
        from qiskit_nature.mappers.second_quantization import JordanWignerMapper
        from qiskit_nature.algorithms import GroundStateEigensolver
    except:
        print("Please install qiskit_nature.")
        return None

    driver = ElectronicStructureMoleculeDriver(
        molecule, basis=basis, driver_type=ElectronicStructureDriverType.PYSCF
    )

    es_problem = ElectronicStructureProblem(driver)
    qubit_converter = QubitConverter(JordanWignerMapper())

    # quantum_instance = QuantumInstance(backend=Aer.get_backend("aer_simulator_statevector"))
    # vqe_solver = VQEUCCFactory(quantum_instance)

    tl_circuit = TwoLocal(
        rotation_blocks=["h", "rx"],
        entanglement_blocks="cz",
        entanglement="full",
        reps=2,
        parameter_prefix="y",
    )

    another_solver = VQE(
        ansatz=tl_circuit,
        quantum_instance=QuantumInstance(Aer.get_backend("aer_simulator"), shots=1024),
        optimizer=COBYLA(maxiter=25),
    )

    calc = GroundStateEigensolver(qubit_converter, another_solver)
    res = calc.solve(es_problem)

    qc = another_solver.get_optimal_circuit()
    qc.name = "groundstate"
    qc.measure_all()

    return qc



from qiskit import QuantumCircuit, Aer, __qiskit_version__
from qiskit.compiler import transpile
from qiskit.transpiler import CouplingMap
from qiskit.visualization import plot_histogram
from qiskit.circuit import qpy_serialization
from qiskit.algorithms import EstimationProblem

from datetime import date

import networkx as nx
import numpy as np
from os import path, mkdir

from qiskit.test.mock import (
    FakeBogota,
    FakeCasablanca,
    FakeGuadalupe,
    FakeMontreal,
    FakeManhattan,
)

qasm_path = "./benchviewer/static/files/qasm_output/"


def set_qasm_output_path(new_path: str = "./benchviewer/static/files/qasm_output/"):
    global qasm_path
    qasm_path = new_path


def get_qasm_output_path():
    """Returns the path where all .qasm files are stored."""
    return qasm_path


def get_compiled_circuit_with_gateset(
    qc: QuantumCircuit,
    opt_level: int = 1,
    basis_gates=None,
    c_map: CouplingMap = None,
):
    """Returns a transpiled quantum circuit according to the given input parameters.

    Keyword arguments:
    qc -- to be transpiled QuantumCircuit
    opt_level -- optimization level
    basis_gates -- used gates of resulting quantum circuit
    c_map -- to be used coupling map
    """

    if basis_gates is None:
        basis_gates = ["id", "rz", "sx", "x", "cx", "reset"]

    t_qc = transpile(
        qc, basis_gates=basis_gates, optimization_level=opt_level, coupling_map=c_map
    )
    return t_qc


def get_rigetti_native_gates():
    return ["rx", "rz", "cz", "id", "reset"]


def save_as_qasm(
    qc: QuantumCircuit,
    filename: str,
    gate_set: list = None,
    opt_level: int = -1,
    mapped: bool = False,
    c_map=None,
    arch_name: str = "",
):
    """Saves a quantum circuit as a qasm file.

    Keyword arguments:
    qc -- to be saved QuantumCircuit
    filename -- filename
    gate_set -- set of used gates
    opt_level -- optimization level
    mapped -- boolean indicating whether the quantum circuit is mapped to a specific hardware layout
    c_map -- coupling map of used hardware layout
    arch_name -- name of used hardware layout
    """

    if c_map is None:
        c_map = []

    qasm_output_folder = get_qasm_output_path()
    print(qasm_output_folder)
    with open(qasm_output_folder + filename + ".qasm", "w") as f:
        f.write("// Benchmark was created by MQT Bench on " + str(date.today()) + "\n")
        f.write(
            "// For more information about MQT Bench, please visit https://www.cda.cit.tum.de/mqtbench/"
            + "\n"
        )
        f.write("// MQT Bench version: " + "0.1.0" + "\n")
        f.write("// Qiskit version: " + str(__qiskit_version__) + "\n")
        if gate_set:
            f.write("// Used Gate Set: " + str(gate_set) + "\n")
        if opt_level >= 0:
            f.write("// Optimization Level: " + str(opt_level) + "\n")
        if mapped:
            f.write("// Coupling List: " + str(c_map) + "\n")
            if arch_name:
                f.write("// Compiled for architecture: " + arch_name + "\n")
        f.write("\n")
        f.write(qc.qasm())
    f.close()


def serialize_qc(qc: QuantumCircuit, filename: str):
    """Saves a quantum circuit as a qpy file.

    Keyword arguments:
    qc -- to be saved QuantumCircuit
    filename -- filename
    """

    with open("qpy_output/" + filename + ".qpy", "wb") as f:
        qpy_serialization.dump(qc, f)
    f.close()


def get_examplary_max_cut_qp(n_nodes: int, degree: int = 2):
    """Returns a quadratic problem formulation of a max cut problem of a random graph.

    Keyword arguments:
    n_nodes -- number of graph nodes (and also number of qubits)
    degree -- edges per node
    """
    try:
        from qiskit_optimization.applications import Maxcut
    except:

        print("Please install qiskit_optimization.")
        return None
    graph = nx.random_regular_graph(d=degree, n=n_nodes, seed=111)
    maxcut = Maxcut(graph)
    return maxcut.to_quadratic_program()


def sim_and_print_hist(qc: QuantumCircuit, filename: str):
    """Simulates a given quantum circuit and prints its resulting histogram.

    Keyword arguments:
    qc -- to be simulated quantum circuit
    filename -- filename of the histogram
    """

    if not path.isdir("./hist_output"):
        mkdir("hist_output")

    simulator = Aer.get_backend("aer_simulator")
    result = simulator.run(qc.decompose(), shots=1024).result()
    counts = result.get_counts()
    plot = plot_histogram(counts, figsize=(15, 5), title=filename)
    plot.savefig("hist_output/" + filename + "_hist" + ".png", bbox_inches="tight")


def save_circ(qc: QuantumCircuit, filename: str):
    """Save schematic picture of a given quantum circuit.

    Keyword arguments:
    qc -- to be saved quantum circuit
    filename -- filename of the schematic picture
    """
    if not path.isdir("./hist_output"):
        mkdir("hist_output")

    qc.draw(output="mpl", filename="hist_output/" + filename + ".png")


class BernoulliA(QuantumCircuit):
    """A circuit representing the Bernoulli A operator."""

    def __init__(self, probability):
        super().__init__(1)  # circuit on 1 qubit

        theta_p = 2 * np.arcsin(np.sqrt(probability))
        self.ry(theta_p, 0)


class BernoulliQ(QuantumCircuit):
    """A circuit representing the Bernoulli Q operator."""

    def __init__(self, probability):
        super().__init__(1)  # circuit on 1 qubit

        self._theta_p = 2 * np.arcsin(np.sqrt(probability))
        self.ry(2 * self._theta_p, 0)

    def power(self, k):
        # implement the efficient power of Q
        q_k = QuantumCircuit(1)
        q_k.ry(2 * k * self._theta_p, 0)
        return q_k


def get_estimation_problem():
    """Returns a estimation problem instance for a fixed p value."""

    p = 0.2

    a = BernoulliA(p)
    q = BernoulliQ(p)

    problem = EstimationProblem(
        state_preparation=a,  # A operator
        grover_operator=q,  # Q operator
        objective_qubits=[
            0
        ],  # the "good" state Psi1 is identified as measuring |1> in qubit 0
    )

    return problem


def get_rigetti_c_map(circles: int = 4):
    """Returns a coupling map of the circular layout scheme used by Rigetti.

    Keyword arguments:
    circles -- number of circles, each one comprises 8 qubits
    """
    if circles != 10:
        c_map_rigetti = []
        for j in range(circles):
            for i in range(0, 7):
                c_map_rigetti.append([i + j * 8, i + 1 + j * 8])

                if i == 6:
                    c_map_rigetti.append([0 + j * 8, 7 + j * 8])

            if j != 0:
                c_map_rigetti.append([j * 8 - 6, j * 8 + 5])
                c_map_rigetti.append([j * 8 - 7, j * 8 + 6])

        inverted = [[item[1], item[0]] for item in c_map_rigetti]
        c_map_rigetti = c_map_rigetti + inverted
    else:
        c_map_rigetti = []
        for j in range(5):
            for i in range(0, 7):
                c_map_rigetti.append([i + j * 8, i + 1 + j * 8])

                if i == 6:
                    c_map_rigetti.append([0 + j * 8, 7 + j * 8])

            if j != 0:
                c_map_rigetti.append([j * 8 - 6, j * 8 + 5])
                c_map_rigetti.append([j * 8 - 7, j * 8 + 6])

        for j in range(5):
            m = 8 * j + 5 * 8
            for i in range(0, 7):
                c_map_rigetti.append([i + m, i + 1 + m])

                if i == 6:
                    c_map_rigetti.append([0 + m, 7 + m])

            if j != 0:
                c_map_rigetti.append([m - 6, m + 5])
                c_map_rigetti.append([m - 7, m + 6])

        for n in range(5):
            c_map_rigetti.append([n * 8 + 3, n * 8 + 5 * 8])
            c_map_rigetti.append([n * 8 + 4, n * 8 + 7 + 5 * 8])

        inverted = [[item[1], item[0]] for item in c_map_rigetti]
        c_map_rigetti = c_map_rigetti + inverted

    return c_map_rigetti


def get_google_c_map():
    """Returns a coupling map of the hardware layout scheme used by Google's Sycamore chip."""
    c_map_google = []
    # iterate through each second line of qubits in sycamore architecture
    for j in range(1, 8, 2):
        for i in range(6):
            # connect qubit to upper left und lower left qubits
            c_map_google.append([i + 6 * j, i + 6 * j + 5])
            c_map_google.append([i + 6 * j, i + 6 * j - 6])

            # as long as it is not the most right qubit: connect it to upper right and lower right qubit
            if i != 5:
                c_map_google.append([i + 6 * j, i + 6 * j - 5])
                c_map_google.append([i + 6 * j, i + 6 * j + 7])

    inverted = [[item[1], item[0]] for item in c_map_google]
    c_map_google = c_map_google + inverted

    return c_map_google


def handle_algorithm_level(
    qc: QuantumCircuit, num_qubits: int, save_png: bool, save_hist: bool
):
    """Handles the creation of the benchmark on the algorithm level.

    Keyword arguments:
    qc -- quantum circuit which shall be created
    num_qubits -- number of qubits
    save_png -- flag indicated whether to save schematic picture of quantum circuit
    save_hist -- flag indicated whether to simulate circuit and save its histogram

    Return values:
    filename_algo -- the filename of the created and saved benchmark
    depth -- circuit depth of created benchmark
    num_qubits -- number of qubits of generated circuit
    """
    filename_algo = qc.name + "_algorithm_" + str(num_qubits)
    if path.isfile("qpy_output/" + filename_algo + ".qpy"):
        filepath = "qpy_output/" + filename_algo + ".qpy"
        print(filepath + " already exists")
        with open(filepath, "rb") as fd:
            qc = qpy_serialization.load(fd)[0]
        depth = qc.depth()
    else:
        serialize_qc(qc, filename_algo)
        if save_png:
            save_circ(qc, filename_algo)
        if save_hist:
            sim_and_print_hist(qc, filename_algo)
        depth = qc.depth()

    return filename_algo, depth, num_qubits


def get_indep_level(
    qc: QuantumCircuit,
    num_qubits: int,
    save_png: bool,
    save_hist: bool,
    file_precheck: bool,
):
    """Handles the creation of the benchmark on the target-independent level.

    Keyword arguments:
    qc -- quantum circuit which the to be created benchmark circuit is based on
    num_qubits -- number of qubits
    save_png -- flag indicated whether to save schematic picture of quantum circuit
    save_hist -- flag indicated whether to simulate circuit and save its histogram
    file_precheck -- flag indicating whether to check whether the file already exists before creating it (again)

    Return values:
    filename_indep -- the filename of the created and saved benchmark
    depth -- circuit depth of created benchmark
    num_qubits -- number of qubits of generated circuit
    """

    filename_indep = qc.name + "_indep_" + str(num_qubits)
    qasm_output_folder = get_qasm_output_path()
    filepath = qasm_output_folder + filename_indep + ".qasm"
    if path.isfile(filepath) and file_precheck:
        print(filepath + " already exists")
        qc = QuantumCircuit.from_qasm_file(filepath)
        depth = qc.depth()
        return filename_indep, depth, num_qubits

    else:
        openqasm_gates = get_openqasm_gates()
        target_independent = transpile(
            qc, basis_gates=openqasm_gates, optimization_level=1
        )
        save_as_qasm(target_independent, filename_indep)
        if save_png:
            save_circ(target_independent, filename_indep)
        if save_hist:
            sim_and_print_hist(target_independent, filename_indep)

        depth = target_independent.depth()
        return filename_indep, depth, qc.num_qubits


def get_native_gates_level(
    qc: QuantumCircuit,
    gate_set: list,
    gate_set_name: str,
    opt_level: int,
    num_qubits: int,
    save_png: bool,
    save_hist: bool,
    file_precheck: bool,
):
    """Handles the creation of the benchmark on the target-dependent native gates level.

    Keyword arguments:
    qc -- quantum circuit which the to be created benchmark circuit is based on
    gate_set -- set of to be used gates
    gate_set_name -- name of this gate set
    opt_level -- optimization level
    num_qubits -- number of qubits
    save_png -- flag indicated whether to save schematic picture of quantum circuit
    save_hist -- flag indicated whether to simulate circuit and save its histogram
    file_precheck -- flag indicating whether to check whether the file already exists before creating it (again)

    Return values:
    filename_nativegates -- the filename of the created and saved benchmark
    depth -- circuit depth of created benchmark
    num_qubits/n_actual -- number of qubits of created benchmark
    """

    filename_nativegates = (
        qc.name
        + "_nativegates_"
        + gate_set_name
        + "_opt"
        + str(opt_level)
        + "_"
        + str(num_qubits)
    )

    qasm_output_folder = get_qasm_output_path()
    if (
        path.isfile(qasm_output_folder + filename_nativegates + ".qasm")
        and file_precheck
    ):
        print(qasm_output_folder + filename_nativegates + ".qasm" + " already exists")
        qc = QuantumCircuit.from_qasm_file(
            qasm_output_folder + filename_nativegates + ".qasm"
        )
        depth = qc.depth()
        return filename_nativegates, depth, num_qubits

    else:
        compiled_without_architecture = get_compiled_circuit_with_gateset(
            qc=qc, opt_level=opt_level, basis_gates=gate_set
        )
        n_actual = compiled_without_architecture.num_qubits
        filename_nativegates = (
            qc.name
            + "_nativegates_"
            + gate_set_name
            + "_opt"
            + str(opt_level)
            + "_"
            + str(n_actual)
        )
        save_as_qasm(
            compiled_without_architecture, filename_nativegates, gate_set, opt_level
        )
        if save_png:
            save_circ(compiled_without_architecture, filename_nativegates)
        if save_hist:
            sim_and_print_hist(compiled_without_architecture, filename_nativegates)

        depth = compiled_without_architecture.depth()
        return filename_nativegates, depth, n_actual


def get_mapped_level(
    qc: QuantumCircuit,
    gate_set: list,
    gate_set_name: str,
    opt_level: int,
    num_qubits: int,
    smallest_fitting_arch: bool,
    save_png: bool,
    save_hist: bool,
    file_precheck: bool,
):
    """Handles the creation of the benchmark on the target-dependent mapped level.

    Keyword arguments:
    qc -- quantum circuit which the to be created benchmark circuit is based on
    gate_set -- set of to be used gates
    gate_set_name -- name of this gate set
    opt_level -- optimization level
    num_qubits -- number of qubits
    smallest_fitting_arch -- flag indicating whether smallest fitting mapping scheme shall be used
    save_png -- flag indicated whether to save schematic picture of quantum circuit
    save_hist -- flag indicated whether to simulate circuit and save its histogram
    file_precheck -- flag indicating whether to check whether the file already exists before creating it (again)

    Return values:
    filename_mapped -- the filename of the created and saved benchmark
    depth -- circuit depth of created benchmark
    num_qubits -- number of qubits of generated circuit
    """

    c_map, backend_name, gate_set_name_mapped, c_map_found = select_c_map(
        gate_set_name, smallest_fitting_arch, num_qubits
    )

    if c_map_found:
        filename_mapped = (
            qc.name
            + "_mapped_"
            + gate_set_name_mapped
            + "_opt"
            + str(opt_level)
            + "_"
            + str(num_qubits)
        )
        qasm_output_folder = get_qasm_output_path()
        if (
            path.isfile(qasm_output_folder + filename_mapped + ".qasm")
            and file_precheck
        ):
            print(qasm_output_folder + filename_mapped + ".qasm" + " already exists")
            qc = QuantumCircuit.from_qasm_file(
                qasm_output_folder + filename_mapped + ".qasm"
            )
            depth = qc.depth()
        else:
            compiled_with_architecture = get_compiled_circuit_with_gateset(
                qc=qc, opt_level=opt_level, basis_gates=gate_set, c_map=c_map
            )
            save_as_qasm(
                compiled_with_architecture,
                filename_mapped,
                gate_set,
                opt_level,
                True,
                c_map,
                gate_set_name_mapped + "-" + backend_name,
            )

            if save_png:
                save_circ(compiled_with_architecture, filename_mapped)
            if save_hist:
                sim_and_print_hist(compiled_with_architecture, filename_mapped)

            depth = compiled_with_architecture.depth()
        return filename_mapped, depth, num_qubits
    else:
        return "", 0, 0


def select_c_map(gate_set_name: str, smallest_fitting_arch: bool, num_qubits: int):
    """Returns a suitable coupling map for input parameters

    Keyword arguments:
    gate_set_name -- name of used gate set
    smallest_fitting_arch -- flag indicating whether smallest fitting mapping scheme shall be used
    num_qubits -- number of qubits

    Return values:
    c_map -- coupling map for input parameters
    backend_name -- name of the hardware layout for the respective coupling map
    gate_set_name_mapped -- combination of gate_set_name and smallest_fitting_arch
    c_map_found -- indicator whether a suitable coupling map has been found
    """
    c_map_found = False
    c_map = []
    backend_name = ""
    gate_set_name_mapped = ""

    if gate_set_name == "rigetti":
        c_map_found = True
        if smallest_fitting_arch:
            if num_qubits <= 8:
                c_map = get_rigetti_c_map(1)
                backend_name = "8 qubits"
            elif num_qubits <= 16:
                c_map = get_rigetti_c_map(2)
                backend_name = "Aspen-3: 16 qubits"
            elif num_qubits <= 32:
                c_map = get_rigetti_c_map(4)
                backend_name = "Aspen-10: 32 qubits"
            elif num_qubits <= 40:
                c_map = get_rigetti_c_map(5)
                backend_name = "Aspen-11: 40 qubits"
            elif num_qubits <= 80:
                c_map = get_rigetti_c_map(10)
                backend_name = "Aspen-M-1: 80 qubits"
            else:
                c_map_found = False
            gate_set_name_mapped = gate_set_name + "-s"

        elif num_qubits <= 80:
            c_map = get_rigetti_c_map(10)
            backend_name = "Aspen-M-1: 80 qubits"
            gate_set_name_mapped = gate_set_name + "-b"
        else:
            c_map_found = False

    elif gate_set_name == "ibm":
        c_map_found = True
        if smallest_fitting_arch:
            if num_qubits <= 5:
                backend = FakeBogota()
                c_map = backend.configuration().coupling_map
                backend_name = backend.name()
            elif num_qubits <= 7:
                backend = FakeCasablanca()
                c_map = backend.configuration().coupling_map
                backend_name = backend.name()
            elif num_qubits <= 16:
                backend = FakeGuadalupe()
                c_map = backend.configuration().coupling_map
                backend_name = backend.name()
            elif num_qubits <= 27:
                backend = FakeMontreal()
                c_map = backend.configuration().coupling_map
                backend_name = backend.name()
            elif num_qubits <= 65:
                backend = FakeManhattan()
                c_map = backend.configuration().coupling_map
                backend_name = backend.name()
            elif num_qubits <= 127:
                c_map = get_cmap_imbq_washington()
                backend_name = "Washington"
            else:
                c_map_found = False
            gate_set_name_mapped = gate_set_name + "-s"

        elif num_qubits <= 127:
            c_map = get_cmap_imbq_washington()
            backend_name = "Washington"
            gate_set_name_mapped = gate_set_name + "-b"
        else:
            c_map_found = False

    if c_map_found:
        return c_map, backend_name, gate_set_name_mapped, c_map_found
    else:
        return None, "", "", False


def get_cmap_imbq_washington():
    """Returns the coupling map of the IBM-Q washington quantum computer."""
    c_map_ibmq_washington = [
        [0, 1],
        [1, 2],
        [2, 3],
        [3, 4],
        [4, 5],
        [5, 6],
        [6, 7],
        [7, 8],
        [0, 14],
        [14, 18],
        [18, 19],
        [19, 20],
        [20, 21],
        [21, 22],
        [4, 15],
        [15, 22],
        [22, 23],
        [23, 24],
        [24, 25],
        [25, 26],
        [8, 16],
        [16, 26],
        [26, 27],
        [27, 28],
        [28, 29],
        [29, 30],
        [30, 31],
        [31, 32],
        [9, 10],
        [10, 11],
        [11, 12],
        [12, 13],
        [12, 17],
        [17, 30],
        [32, 36],
        [36, 51],
        [20, 33],
        [33, 39],
        [24, 34],
        [34, 43],
        [28, 35],
        [35, 47],
        [37, 38],
        [38, 39],
        [39, 40],
        [40, 41],
        [41, 42],
        [42, 43],
        [43, 44],
        [44, 45],
        [45, 46],
        [46, 47],
        [47, 48],
        [48, 49],
        [49, 50],
        [50, 51],
        [37, 52],
        [52, 56],
        [41, 53],
        [53, 60],
        [45, 54],
        [54, 64],
        [49, 55],
        [55, 68],
        [56, 57],
        [57, 58],
        [58, 59],
        [59, 60],
        [60, 61],
        [61, 62],
        [62, 63],
        [63, 64],
        [64, 65],
        [65, 66],
        [66, 67],
        [67, 68],
        [68, 69],
        [69, 70],
        [70, 74],
        [74, 89],
        [58, 71],
        [71, 77],
        [62, 72],
        [72, 81],
        [66, 73],
        [73, 85],
        [75, 76],
        [76, 77],
        [77, 78],
        [78, 79],
        [79, 80],
        [80, 81],
        [81, 82],
        [82, 83],
        [83, 84],
        [84, 85],
        [85, 86],
        [86, 87],
        [87, 88],
        [88, 89],
        [75, 90],
        [90, 94],
        [79, 91],
        [91, 98],
        [83, 92],
        [92, 102],
        [87, 93],
        [93, 106],
        [94, 95],
        [95, 96],
        [96, 97],
        [97, 98],
        [98, 99],
        [99, 100],
        [100, 101],
        [101, 102],
        [102, 103],
        [103, 104],
        [104, 105],
        [105, 106],
        [106, 107],
        [107, 108],
        [108, 112],
        [112, 126],
        [96, 109],
        [100, 110],
        [110, 118],
        [104, 111],
        [111, 122],
        [113, 114],
        [114, 115],
        [115, 116],
        [116, 117],
        [117, 118],
        [118, 119],
        [119, 120],
        [120, 121],
        [121, 122],
        [122, 123],
        [123, 124],
        [124, 125],
        [125, 126],
    ]

    inverted = [[item[1], item[0]] for item in c_map_ibmq_washington]
    c_map_ibmq_washington = c_map_ibmq_washington + inverted
    return c_map_ibmq_washington


def get_openqasm_gates():
    """Returns a list of all quantum gates within the openQASM 2.0 standard header."""
    # according to https://github.com/Qiskit/qiskit-terra/blob/main/qiskit/qasm/libs/qelib1.inc
    gate_list = [
        "u3",
        "u2",
        "u1",
        "cx",
        "id",
        "u0",
        "u",
        "p",
        "x",
        "y",
        "z",
        "h",
        "s",
        "sdg",
        "t",
        "tdg",
        "rx",
        "ry",
        "rz",
        "sx",
        "sxdg",
        "cz",
        "cy",
        "swap",
        "ch",
        "ccx",
        "cswap",
        "crx",
        "cry",
        "crz",
        "cu1",
        "cp",
        "cu3",
        "csx",
        "cu",
        "rxx",
        "rzz",
        "rccx",
        "rc3x",
        "c3x",
        "c3sqrtx",
        "c4x",
    ]
    return gate_list


def get_molecule(benchmark_instance_name: str):
    """Returns a Molecule object depending on the parameter value."""
    try:
        from qiskit_nature.drivers import Molecule
    except:
        print("Please install qiskit_nature.")
        return None
    m_1 = Molecule(
        geometry=[["Li", [0.0, 0.0, 0.0]], ["H", [0.0, 0.0, 2.5]]],
        charge=0,
        multiplicity=1,
    )
    m_2 = Molecule(
        geometry=[["H", [0.0, 0.0, 0.0]], ["H", [0.0, 0.0, 0.735]]],
        charge=0,
        multiplicity=1,
    )
    m_3 = Molecule(
        geometry=[
            ["O", [0.0, 0.0, 0.0]],
            ["H", [0.586, 0.757, 0.0]],
            ["H", [0.586, -0.757, 0.0]],
        ],
        charge=0,
        multiplicity=1,
    )
    instances = {"small": m_1, "medium": m_2, "large": m_3}

    return instances[benchmark_instance_name]



# Code from https://qiskit.org/documentation/nature/tutorials/04_excited_states_solvers.html


from qiskit import Aer
from qiskit.utils import QuantumInstance
from qiskit.circuit.library import TwoLocal
from qiskit.algorithms import VQE
from qiskit.algorithms.optimizers import SLSQP, COBYLA


def create_circuit(molecule, basis: str = "sto3g"):
    """Returns a quantum circuit implementing Excited State Estimation.

    Keyword arguments:
    molecule -- molecule for which the excited state shall be estimated.
    basis -- basis used for estimation
    """
    try:
        from qiskit_nature.drivers import Molecule
        from qiskit_nature.drivers.second_quantization import (
            ElectronicStructureDriverType,
            ElectronicStructureMoleculeDriver,
        )
        from qiskit_nature.problems.second_quantization import (
            ElectronicStructureProblem,
        )
        from qiskit_nature.converters.second_quantization import QubitConverter
        from qiskit_nature.mappers.second_quantization import JordanWignerMapper
        from qiskit_nature.algorithms import GroundStateEigensolver, QEOM
    except:
        print("Please install qiskit_nature.")
        return None

    driver = ElectronicStructureMoleculeDriver(
        molecule, basis=basis, driver_type=ElectronicStructureDriverType.PYSCF
    )

    es_problem = ElectronicStructureProblem(driver)
    qubit_converter = QubitConverter(JordanWignerMapper())

    # This first part sets the ground state solver
    # see more about this part in the ground state calculation tutorial
    # quantum_instance = QuantumInstance(Aer.get_backend("aer_simulator_statevector"))
    # solver = VQEUCCFactory(quantum_instance)

    tl_circuit = TwoLocal(
        rotation_blocks=["h", "rx"],
        entanglement_blocks="cz",
        entanglement="full",
        reps=2,
        parameter_prefix="y",
    )

    another_solver = VQE(
        ansatz=tl_circuit,
        quantum_instance=QuantumInstance(Aer.get_backend("aer_simulator"), shots=1024),
        optimizer=COBYLA(maxiter=25),
    )
    gsc = GroundStateEigensolver(qubit_converter, another_solver)

    # The qEOM algorithm is simply instantiated with the chosen ground state solver
    qeom_excited_states_calculation = QEOM(gsc, "sd")

    qeom_results = qeom_excited_states_calculation.solve(es_problem)

    qc = another_solver.get_optimal_circuit()
    qc.name = "excitedstate"
    qc.measure_all()

    # print(qeom_results)
    return qc



from setuptools import setup
import os

README_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")
with open(README_PATH) as readme_file:
    README = readme_file.read()

setup(
    name="mqt.bench",
    packages=[
        "mqt.bench",
        "mqt.bench.utils",
        "mqt.bench.benchmarks",
        "mqt.bench.benchmarks.qiskit_application_finance",
        "mqt.bench.benchmarks.qiskit_application_ml",
        "mqt.bench.benchmarks.qiskit_application_optimization",
        "mqt.bench.benchmarks.qiskit_application_nature",
    ],
    version="0.1.0rc1",
    python_requires=">=3.8",
    license="MIT",
    description="MQT Bench - A MQT tool for Benchmarking Quantum Software Tools",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Nils Quetschlich",
    author_email="nils.quetschlich@tum.de",
    url="https://github.com/cda-tum/mqtbench",
    keywords="mqt quantum benchmarking performance testing",
    install_requires=[
        "matplotlib~=3.5.1",
        "qiskit~=0.35",
        "pandas~=1.3.5",
        "flask~=2.0.2",
        "networkx~=2.7.1",
        "pytest~=7.1.1",
        "pylatexenc~=2.10",
    ],
    extras_require={
        "all": [
            "qiskit_finance~=0.3.1",
            "qiskit_machine-learning==0.3.1",
            "qiskit_nature~=0.3.1",
            "qiskit_optimization~=0.3.2",
            "qiskit-aqua~=0.9.5",
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
    ],
)



# Code from https://qiskit.org/documentation/tutorials/optimization/7_examples_vehicle_routing.html

import numpy as np

from qiskit import Aer
from qiskit.utils import QuantumInstance, algorithm_globals
from qiskit.algorithms import VQE
from qiskit.algorithms.optimizers import SLSQP


class Initializer:
    def __init__(self, n):
        self.n = n

    def generate_instance(self):
        n = self.n
        np.random.seed = 10

        xc = (np.random.rand(n) - 0.5) * 10
        yc = (np.random.rand(n) - 0.5) * 10

        instance = np.zeros([n, n])
        for ii in range(0, n):
            for jj in range(ii + 1, n):
                instance[ii, jj] = (xc[ii] - xc[jj]) ** 2 + (yc[ii] - yc[jj]) ** 2
                instance[jj, ii] = instance[ii, jj]

        return xc, yc, instance


class QuantumOptimizer:
    def __init__(self, instance, n, k):

        self.instance = instance
        self.n = n
        self.K = k

    def binary_representation(self, x_sol=0):

        instance = self.instance
        n = self.n
        k = self.K

        a = np.max(instance) * 100  # A parameter of cost function

        # Determine the weights w
        instance_vec = instance.reshape(n**2)
        w_list = [instance_vec[x] for x in range(n**2) if instance_vec[x] > 0]
        w = np.zeros(n * (n - 1))
        for ii in range(len(w_list)):
            w[ii] = w_list[ii]

        id_n = np.eye(n)
        im_n_1 = np.ones([n - 1, n - 1])
        iv_n_1 = np.ones(n)
        iv_n_1[0] = 0
        iv_n = np.ones(n - 1)
        neg_iv_n_1 = np.ones(n) - iv_n_1

        v = np.zeros([n, n * (n - 1)])
        for ii in range(n):
            count = ii - 1
            for jj in range(n * (n - 1)):

                if jj // (n - 1) == ii:
                    count = ii

                if jj // (n - 1) != ii and jj % (n - 1) == count:
                    v[ii][jj] = 1.0

        vn = np.sum(v[1:], axis=0)

        # Q defines the interactions between variables
        q = a * (np.kron(id_n, im_n_1) + np.dot(v.T, v))

        # g defines the contribution from the individual variables
        g = (
            w
            - 2 * a * (np.kron(iv_n_1, iv_n) + vn.T)
            - 2 * a * k * (np.kron(neg_iv_n_1, iv_n) + v[0].T)
        )

        # c is the constant offset
        c = 2 * a * (n - 1) + 2 * a * (k**2)

        try:
            max(x_sol)
            # Evaluates the cost distance from a binary representation of a path
            fun = (
                lambda x: np.dot(np.around(x), np.dot(q, np.around(x)))
                + np.dot(g, np.around(x))
                + c
            )
            cost = fun(x_sol)
        except:
            cost = 0

        return q, g, c, cost

    def construct_problem(self, q, g, c):
        try:
            from qiskit_optimization import QuadraticProgram
        except:
            print("Please install qiskit_optimization.")
            return None

        qp = QuadraticProgram()
        for i in range(self.n * (self.n - 1)):
            qp.binary_var(str(i))
        qp.objective.quadratic = q
        qp.objective.linear = g
        qp.objective.constant = c
        return qp

    def solve_problem(self, qp):

        try:
            from qiskit_optimization.algorithms import MinimumEigenOptimizer
        except:
            print("Please install qiskit_optimization.")
            return None

        algorithm_globals.random_seed = 10
        quantum_instance = QuantumInstance(
            Aer.get_backend("aer_simulator"),
            seed_simulator=algorithm_globals.random_seed,
            seed_transpiler=algorithm_globals.random_seed,
            shots=1024,
        )

        vqe = VQE(quantum_instance=quantum_instance, optimizer=SLSQP(maxiter=25))
        optimizer = MinimumEigenOptimizer(min_eigen_solver=vqe)
        result = optimizer.solve(qp)
        # compute cost of the obtained result
        _, _, _, level = self.binary_representation(x_sol=result.x)
        return result.x, level, vqe


def create_circuit(num_nodes: int = 3, num_vehs: int = 2):
    """Returns a quantum circuit solving a routing problem.

    Keyword arguments:
    num_nodes -- number of to be visited nodes
    num_vehs -- number of used vehicles
    """

    # Initialize the problem by defining the parameters
    n = num_nodes  # number of nodes + depot (n+1)
    k = num_vehs  # number of vehicles
    # Initialize the problem by randomly generating the instance
    initializer = Initializer(n)
    xc, yc, instance = initializer.generate_instance()

    quantum_optimizer = QuantumOptimizer(instance, n, k)
    q, g, c, binary_cost = quantum_optimizer.binary_representation()
    qp = quantum_optimizer.construct_problem(q, g, c)
    # Instantiate the quantum optimizer class with parameters:
    quantum_solution, quantum_cost, vqe = quantum_optimizer.solve_problem(qp)

    vqe_result = vqe.compute_minimum_eigenvalue(qp.to_ising()[0])
    qc = vqe.ansatz.bind_parameters(vqe_result.optimal_point)

    qc.measure_all()
    qc.name = "routing"

    return qc



# Code based on https://qiskit.org/documentation/tutorials/finance/04_european_put_option_pricing.html

import numpy as np
from qiskit.algorithms import IterativeAmplitudeEstimation, EstimationProblem
from qiskit.circuit.library import LinearAmplitudeFunction


def create_circuit(num_uncertainty_qubits: int = 5):
    """Returns a quantum circuit of Iterative Amplitude Estimation applied to a problem instance of pricing put options.

    Keyword arguments:
    num_uncertainty_qubits -- number of qubits to measure uncertainty
    """

    try:
        from qiskit_finance.circuit.library import LogNormalDistribution
    except:
        print("Please install qiskit_finance.")
        return None

    num_uncertainty_qubits = num_uncertainty_qubits

    # parameters for considered random distribution
    s = 2.0  # initial spot price
    vol = 0.4  # volatility of 40%
    r = 0.05  # annual interest rate of 4%
    t = 40 / 365  # 40 days to maturity

    mu = (r - 0.5 * vol**2) * t + np.log(s)
    sigma = vol * np.sqrt(t)
    mean = np.exp(mu + sigma**2 / 2)
    variance = (np.exp(sigma**2) - 1) * np.exp(2 * mu + sigma**2)
    stddev = np.sqrt(variance)

    # lowest and highest value considered for the spot price; in between, an equidistant discretization is considered.
    low = np.maximum(0, mean - 3 * stddev)
    high = mean + 3 * stddev

    # construct A operator for QAE for the payoff function by
    # composing the uncertainty model and the objective
    uncertainty_model = LogNormalDistribution(
        num_uncertainty_qubits, mu=mu, sigma=sigma**2, bounds=(low, high)
    )

    # set the strike price (should be within the low and the high value of the uncertainty)
    strike_price = 2.126

    # set the approximation scaling for the payoff function
    rescaling_factor = 0.25

    # setup piecewise linear objective fcuntion
    breakpoints = [low, strike_price]
    slopes = [-1, 0]
    offsets = [strike_price - low, 0]
    f_min = 0
    f_max = strike_price - low
    european_put_objective = LinearAmplitudeFunction(
        num_uncertainty_qubits,
        slopes,
        offsets,
        domain=(low, high),
        image=(f_min, f_max),
        breakpoints=breakpoints,
        rescaling_factor=rescaling_factor,
    )

    # construct A operator for QAE for the payoff function by
    # composing the uncertainty model and the objective
    european_put = european_put_objective.compose(uncertainty_model, front=True)

    # set target precision and confidence level
    epsilon = 0.01
    alpha = 0.05

    problem = EstimationProblem(
        state_preparation=european_put,
        objective_qubits=[num_uncertainty_qubits],
        post_processing=european_put_objective.post_processing,
    )

    iae = IterativeAmplitudeEstimation(epsilon, alpha=alpha)
    qc = iae.construct_circuit(problem)

    qc.measure_all()
    qc.name = "pricingput"

    return qc



# Code from https://qiskit.org/documentation/optimization/tutorials/06_examples_max_cut_and_tsp.html

from qiskit import Aer
from qiskit.circuit.library import TwoLocal
from qiskit.algorithms import VQE
from qiskit.algorithms.optimizers import SPSA
from qiskit.utils import algorithm_globals, QuantumInstance


def create_circuit(num_nodes: int):
    """Returns a quantum circuit solving the Travelling Salesman Problem (TSP).

    Keyword arguments:
    num_nodes -- number of to be visited nodes
    """

    try:
        from qiskit_optimization.algorithms import MinimumEigenOptimizer
        from qiskit_optimization.converters import QuadraticProgramToQubo
        from qiskit_optimization.applications import Tsp
    except:
        print("Please install qiskit_optimization.")
        return None

    # Generating a graph of 3 nodes
    n = num_nodes
    tsp = Tsp.create_random_instance(n, seed=10)

    qp = tsp.to_quadratic_program()

    qp2qubo = QuadraticProgramToQubo()
    qubo = qp2qubo.convert(qp)
    qubit_op, offset = qubo.to_ising()

    algorithm_globals.random_seed = 10
    seed = 10
    backend = Aer.get_backend("aer_simulator")
    quantum_instance = QuantumInstance(
        backend, seed_simulator=seed, seed_transpiler=seed, shots=1024
    )

    spsa = SPSA(maxiter=25)
    ry = TwoLocal(qubit_op.num_qubits, "ry", "cz", reps=5, entanglement="linear")
    vqe = VQE(ry, optimizer=spsa, quantum_instance=quantum_instance)

    vqe_result = vqe.compute_minimum_eigenvalue(qubit_op)
    qc = vqe.ansatz.bind_parameters(vqe_result.optimal_point)
    qc.measure_all()
    qc.name = "tsp"

    return qc



# Code based on https://qiskit.org/documentation/tutorials/finance/01_portfolio_optimization.html

from qiskit import Aer
from qiskit.utils import QuantumInstance
from qiskit.algorithms import QAOA
from qiskit.algorithms.optimizers import COBYLA
import datetime


def create_circuit(num_qubits: int):
    """Returns a quantum circuit of QAOA applied to a specific portfolio optimization task.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    """
    try:
        from qiskit_finance.applications import PortfolioOptimization
        from qiskit_finance.data_providers import RandomDataProvider

    except:
        print("Please install qiskit_finance.")
        return None

    try:
        from qiskit_optimization.converters import QuadraticProgramToQubo
    except:
        print("Please install qiskit_optimization.")
        return None

    # set number of assets (= number of qubits)
    num_assets = num_qubits

    # Generate expected return and covariance matrix from (random) time-series
    stocks = [("TICKER%s" % i) for i in range(num_assets)]
    data = RandomDataProvider(
        tickers=stocks,
        start=datetime.datetime(2016, 1, 1),
        end=datetime.datetime(2016, 1, 30),
    )
    data.run()
    mu = data.get_period_return_mean_vector()
    sigma = data.get_period_return_covariance_matrix()

    q = 0.5  # set risk factor
    budget = num_assets // 2  # set budget
    penalty = num_assets  # set parameter to scale the budget penalty term

    portfolio = PortfolioOptimization(
        expected_returns=mu, covariances=sigma, risk_factor=q, budget=budget
    )
    qp = portfolio.to_quadratic_program()
    conv = QuadraticProgramToQubo()
    qp_qubo = conv.convert(qp)

    backend = Aer.get_backend("aer_simulator")
    seed = 10

    cobyla = COBYLA()
    cobyla.set_options(maxiter=25)

    sim = QuantumInstance(
        backend=backend, seed_simulator=seed, seed_transpiler=seed, shots=1024
    )

    qaoa = QAOA(cobyla, 3, quantum_instance=sim)
    qaoa.random_seed = seed
    qaoa_result = qaoa.compute_minimum_eigenvalue(qp_qubo.to_ising()[0])
    qc = qaoa.ansatz.bind_parameters(qaoa_result.optimal_point)

    qc.name = "portfolioqaoa"
    qc.measure_all()

    return qc



from qiskit import QuantumCircuit, QuantumRegister, AncillaRegister


def create_circuit(
    n: int,
    depth: int = 3,
    coin_state_preparation: QuantumCircuit = None,
    ancillary_mode: str = "noancilla",
):
    """Returns a quantum circuit implementing the Quantum Walk algorithm.

    Keyword arguments:
    num_qubits -- number of qubits of the returned quantum circuit
    depth -- number of quantum steps
    coin_state_preparation -- optional quantum circuit for state preparation
    ancillary_mode -- defining the decomposition scheme
    """

    n = n - 1  # because one qubit is needed for the coin
    coin = QuantumRegister(1, "coin")
    node = QuantumRegister(n, "node")

    n_anc = 0
    if ancillary_mode == "recursion" and n > 3:
        n_anc = 1
    if (ancillary_mode == "v-chain" or ancillary_mode == "v-chain-dirty") and n > 2:
        n_anc = n - 2

    if n_anc == 0:
        qc = QuantumCircuit(node, coin, name="qwalk")

        # coin state preparation
        if coin_state_preparation is not None:
            qc.append(coin_state_preparation, coin[:])

        for _ in range(depth):
            # Hadamard coin operator
            qc.h(coin)

            # controlled increment
            for i in range(0, n - 1):
                qc.mcx(coin[:] + node[i + 1 :], node[i], mode=ancillary_mode)
            qc.cx(coin, node[n - 1])

            # controlled decrement
            qc.x(coin)
            qc.x(node[1:])
            for i in range(0, n - 1):
                qc.mcx(coin[:] + node[i + 1 :], node[i], mode=ancillary_mode)
            qc.cx(coin, node[n - 1])
            qc.x(node[1:])
            qc.x(coin)
    else:
        anc = AncillaRegister(n_anc, "anc")
        qc = QuantumCircuit(node, coin, anc, name="qwalk")

        # coin state preparation
        if coin_state_preparation is not None:
            qc.append(coin_state_preparation, coin[:])

        for _ in range(depth):
            # Hadamard coin operator
            qc.h(coin)

            # controlled increment
            for i in range(0, n - 1):
                qc.mcx(
                    coin[:] + node[i + 1 :],
                    node[i],
                    mode=ancillary_mode,
                    ancilla_qubits=anc[:],
                )
            qc.cx(coin, node[n - 1])

            # controlled decrement
            qc.x(coin)
            qc.x(node[1:])
            for i in range(0, n - 1):
                qc.mcx(
                    coin[:] + node[i + 1 :],
                    node[i],
                    mode=ancillary_mode,
                    ancilla_qubits=anc[:],
                )
            qc.cx(coin, node[n - 1])
            qc.x(node[1:])
            qc.x(coin)

    qc.measure_all()

    return qc
